
from LoG import *
from pathlib import Path
import orgparse
import re
import importer_common as ic

NAME="Import Org-Mode (Experimental)"
VERSION="0.1"
CATEGORY=SCtg.IMPORTER


def run(path):
    _import_org_file(path)

    refresh_panels()

    return True

_ORDERED_RE = re.compile(r"^( *)(\d+)[.)]\s")
_UNORDERED_RE = re.compile(r"^( *)([-+*])\s")
_TODO_RE = re.compile(r"^\[([ \-Xx])\]\s")
_QUOTE_RE = re.compile(r"^ *?#\+(BEGIN|END)_([A-Z]+)( (\w+))?")
_HRULE_RE = re.compile(r'^[ \t]*-{5,}[ \t]*$')
_TABLE_ROW_RE = re.compile(r'^[ \t]*\|')
_CAPTION_RE = re.compile(r'^[ \t]*#\+CAPTION:[ \t]*(.+)$', re.IGNORECASE)
_NAME_RE = re.compile(r'^[ \t]*#\+NAME:[ \t]*(.+)$', re.IGNORECASE)

# def _join_para_text(text: str) -> str:
#     text = re.sub(r"\s+", " ", text)
#     return text

# FORMATTING =======================================================================================
EMPH_FMT = {
    '*': VT.FMT.BOLD,
    '/': VT.FMT.ITALIC,
    '_': VT.FMT.UNDERLINE,
    '+': VT.FMT.STRIKETHRU,
}
LITERAL_FMT = {
    '~': VT.FMT.HIGHLIGHT,   # inline code
    '=': VT.FMT.FADED,       # verbatim
}

_ANY_EMPH_RE = re.compile(
    r'(?:(?<=^)|(?<=[ \t([{\'"-]))'
    r'([*/_+=~])'
    r'([^\s]|[^\s].*?[^\s])'
    r'\1'
    r'(?=[ \t.,:!?;\'")\]}-]|$)',
    re.MULTILINE
)

def _parse_emphasis(text):
    out = []
    spans = []
    pos = 0
    while True:
        m = _ANY_EMPH_RE.search(text, pos)
        if not m:
            out.append(text[pos:])
            break

        out.append(text[pos:m.start()])
        marker, inner = m.group(1), m.group(2)
        start = sum(len(s) for s in out)

        if marker in LITERAL_FMT:
            out.append(inner)
            spans.append((LITERAL_FMT[marker], start, start + len(inner)))
        else:
            inner_plain, inner_spans = _parse_emphasis(inner)
            for f, s, e in inner_spans:
                spans.append((f, s + start, e + start))
            out.append(inner_plain)
            spans.append((EMPH_FMT[marker], start, start + len(inner_plain)))

        pos = m.end()

    return ''.join(out), spans

# PARA LEVEL FORMATTING ============================================================================
def _infer_para_type(text):
    ps = None
    indent_level = 0
    quote_type = None
    F_quot_bgn = False

    if (m := _ORDERED_RE.match(text)):
        ps = VT.PLS.NUMBER
        print( "number ind level: ", text)
        text = text[m.end():].strip()
        indent_level = len(m.group(1)) // 2

    elif (m := _UNORDERED_RE.match(text) ):
        ps = VT.PLS.BULLET
        print( "bullet ind level: ", text)
        text = text[m.end():].strip()
        indent_level = len(m.group(1)) // 2

        # sub-style:
        m = _TODO_RE.match(text)
        if m:
            match m.group(1):
                case ' ':
                    ps = VT.PLS.TODO
                case '+':
                    ps = VT.PLS.PROGRESSED
                case 'x' | 'X':
                    ps = VT.PLS.DONE
            text = text[m.end():].strip()

    elif (m := _QUOTE_RE.match(text)):
        match m.group(2):
            case 'QUOTE':
                quote_type = VT.QT.LITERARY
                F_quot_bgn = m.group(1) == "BEGIN"
            case 'VERSE' | 'EXAMPLE':
                quote_type = VT.QT.GENERIC
                F_quot_bgn = m.group(1) == "BEGIN"
            case 'SRC':
                quote_type = ic.get_code_from_str(m.group(4)) if m.group(4) else VT.QT.GENERIC
                if quote_type:
                    F_quot_bgn = m.group(1) == "BEGIN"
        text = ""

    elif ( m := _TABLE_ROW_RE.match(text)):
        quote_type = VT.QT.GENERIC
        F_quot_bgn = True

    elif ( m := _CAPTION_RE.match(text)):
        text = m.group(1).strip()

    elif ( m := _NAME_RE.match(text)):
        text = ""

    return text, ps, indent_level, quote_type, F_quot_bgn

def _import_node(entry, para_last, node):
    # Add heading paragraph
    heading = entry.append_paragraph(node.heading)

    # Heading levels
    match node.level:
        case 1:
            heading.set_heading_level(VT.PHS.LARGE)
            heading.set_hrule(True)
        case 2:
            heading.set_heading_level(VT.PHS.LARGE)
        case 3:
            heading.set_heading_level(VT.PHS.MEDIUM)
        case 4 | 5:
            heading.add_format(VT.FMT.BOLD)

    para_last = heading

    # Add node body
    body = node.body

    para_text_cum = ""
    pls_last = None # para list style
    ind_lvl_last = 0
    qt_last = None # quotation type
    text_stripped = ""

    def _consume_text():
        nonlocal para_text_cum
        if para_text_cum:
            nonlocal para_last
            nonlocal pls_last
            nonlocal ind_lvl_last

            if qt_last is None or qt_last == VT.QT.LITERARY:
                para_text_cum = para_text_cum.strip()
            plain_text, spans = _parse_emphasis(para_text_cum)
            para_last = entry.append_paragraph(plain_text)

            # para level formatting:
            if pls_last:
                para_last.set_list_type(pls_last)
                para_last.set_indent_level(ind_lvl_last)
            if qt_last:
                para_last.set_quot_type(qt_last)

            # rich text formatting:
            for fmt, bgn, end in spans:
                para_last.add_format(fmt, bgn, end)

            para_text_cum = "" # cumulative text is consumed
        pls_last = pls
        ind_lvl_last = indent_level

    if body:
        for para_text in body.split("\n"):
            if not para_text.strip():
                pls = None # set new style to be set in _consume_text() to None
                indent_level = 0 # set ind lvl to be set in _consume_text() to 0
                _consume_text()
                continue

            if _HRULE_RE.match(para_text):
                pls = None # set new style to be set in _consume_text() to None
                indent_level = 0 # set ind lvl to be set in _consume_text() to 0
                _consume_text()
                para_last = entry.append_paragraph("")
                para_last.set_hrule(True)
                continue


            text_stripped, pls, indent_level, qt, F_qt_bgn = _infer_para_type(para_text)

            if pls != None or qt != None or (qt_last != None and qt_last != VT.QT.LITERARY):
                # flush buffer if current para is a list item or quote delimiter
                # or we are in a code block
                # TODO: add a special case for VERSE which is regarded like code but should be formatted like
                # literary quote
                _consume_text()
                if qt != None:
                    if F_qt_bgn and qt_last != None:
                        print("Nesting quotations is not supported!")
                    qt_last = qt if F_qt_bgn else None 
            if text_stripped:
                para_text_cum += " " + text_stripped

        # Remainder:
        _consume_text()

    # Process children recursively
    for child in node.children:
        para_last = _import_node(
            entry,
            para_last,
            child
        )

    return para_last

def _get_org_title(filepath):
    with open(filepath, encoding="utf-8") as f:
        for line in f:
            line = line.strip()

            if line.upper().startswith("#+TITLE:"):
                return line[8:].strip()

    return Path(filepath).stem

def _import_org_file(filepath):
    prev_para = None
    root = orgparse.load(filepath)
    title = _get_org_title(filepath)
    entry = Diary.get_main().create_entry(VT.ETS.NAME_ONLY, title )

    # Import all top-level headings
    for node in root.children:
        prev_para = _import_node(
            entry,
            prev_para,
            node
        )




# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *

NAME="Move Paragraphs into Child Entry"
VERSION="0.1"
CATEGORY=SCtg.PARA_MODIFIER

def run(paras):
    # get the parent entry i.e. the current host of the selected paragraphs:
    entry_p = paras[0].get_host()
    #create a child entry:
    entry_c = Diary.get_main().create_entry_child(entry_p)
    # move the pragraphs:
    for p in paras:
        p.move_to_entry(entry_c)
    # remove the empty paragraph added in create_entry_rel:
    entry_c.remove_paragraphs(0)
    entry_p.set_expanded(True)
    refresh_panels()



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *

NAME="Move under a Heading"
VERSION="0.1"
CATEGORY=SCtg.WORD_MODIFIER

def run(para, pos_bgn, pos_end):
    #check the heading level:
    if para.get_heading_level() != VT.PHS.NORMAL_TEXT:
        show_message("Headings cannot be moved!")
        return

    # get headings in the host entry:
    headings = []
    for p in para.get_host():
        heading_lvl = p.get_heading_level()
        if heading_lvl == VT.PHS.LARGE or heading_lvl == VT.PHS.MEDIUM:
            if p != _get_parent_heading(para): # omit current parent
                headings.append((p.get_text(), p.get_id()))

    fields = [ { "type": FT.OPTION, "name": "Heading", "opts": headings },
               { "type": FT.HIDDEN, "name": "Para",    "value": para.get_id() } ]
    # hidden field is used to pass the para-to-move to the input handler

    get_user_input( "Select the heading to move under", NAME, "handle_ret", fields)

# handle user input:
def handle_ret(user_info : Dict[str, Any]):
    diary = Diary.get_main()
    p_heading = diary.get_paragraph_by_id(user_info['Heading'])
    p_to_move = diary.get_paragraph_by_id(user_info['Para'])
    para_sub_last = _get_last_subpara(p_heading)

    # move:
    p_to_move.move_to_entry( p_heading.get_host(), para_sub_last)

    refresh_panels()

# calculate the insertion point:
def _get_last_subpara(p_heading: Paragraph):
    heading_lvl_last = heading_lvl_orig = p_heading.get_heading_level()
    para_sub_last = p_heading
    para_next = p_heading.get_next()
    while para_next:
        heading_lvl_next = para_next.get_heading_level()
        if heading_lvl_next > heading_lvl_last or heading_lvl_next == heading_lvl_orig:
            break
        para_sub_last = para_next
        heading_lvl_last = heading_lvl_next
        para_next = para_next.get_next()
    return para_sub_last

def _get_parent_heading(para: Paragraph):
    para_prev = para.get_prev()
    while para_prev:
        if para_prev.get_heading_level() > para.get_heading_level():
            return para_prev
        para_prev = para_prev.get_prev()
    return None




# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *

VERSION="0.1"

def get_code_from_str(code_str):
    CODE_LANG_MAP = {
        None: VT.QT.GENERIC,
        "": VT.QT.GENERIC,
        "text": VT.QT.GENERIC,
        "plain": VT.QT.GENERIC,

        "sh": VT.QT.BASH,
        "bash": VT.QT.BASH,

        "c": VT.QT.CPP,
        "cpp": VT.QT.CPP,
        "c++": VT.QT.CPP,

        "fortran": VT.QT.FORTRAN,

        "go": VT.QT.GO,
        "golang": VT.QT.GO,

        "haskell": VT.QT.HASKELL,
        "hs": VT.QT.HASKELL,

        "html": VT.QT.HTML,

        "java": VT.QT.JAVA,

        "js": VT.QT.JAVASCRIPT,
        "javascript": VT.QT.JAVASCRIPT,

        "kotlin": VT.QT.KOTLIN,
        "kt": VT.QT.KOTLIN,

        "lisp": VT.QT.LISP,

        "lua": VT.QT.LUA,

        "pascal": VT.QT.PASCAL,

        "perl": VT.QT.PERL,

        "py": VT.QT.PYTHON,
        "python": VT.QT.PYTHON,

        "ruby": VT.QT.RUBY,

        "rust": VT.QT.RUST,

        "scala": VT.QT.SCALA,

        "sql": VT.QT.SQL,
    }

    return CODE_LANG_MAP.get(code_str.strip().lower(), VT.QT.GENERIC)


def is_boundary_alnum_before(text, i):
    before = text[i - 1] if i > 0 else " "
    return before.isalnum()

def is_boundary_alnum_after(text, i, length):
    after = text[i + length] if i + length < len(text) else " "
    return after.isalnum()



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import json
import datetime
import markdown_parsing as mdp

NAME="Claude History Json (Experimental)"
VERSION="0.1"
CATEGORY=SCtg.IMPORTER

_NOW = Date.get_now()

def run(path):
    dm = Diary.get_main()

    with open(path, "r", encoding="utf-8") as f:
        conversations = json.load(f)

    entry_root = dm.create_entry_rel(None, True, "Claude Import")
    entry_root.append_paragraph(f"Imported on {Date.format_str(_NOW, 'F, T')}")

    for conv in conversations:
        title = conv.get("name") or ""
        if not title: #skip empty conversatons
            continue

        messages = conv.get("chat_messages", {})
        date_str = conv.get("created_at") or ""
        uuid = conv.get("uuid", "")

        _import_entry(dm, entry_root, title, _iso_to_datev(date_str), messages, uuid)

    show_message("Import complete!")
    refresh_panels()

    return True

# HELPERS
def _iso_to_datev(iso_date):
    if iso_date:
        try:
            dt = datetime.datetime.fromisoformat(iso_date)
        except (ValueError, TypeError):
            pass
        else:
            return Date.make(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
    return _NOW

def _import_entry(diary, entry_root, title, date_c, messages, uuid):
    entry = diary.create_entry_rel(entry_root, True, title)
    entry.set_creation_date(date_c)
    para = entry.append_paragraph(uuid)
    para.add_format(VT.FMT.FADED)

    for msg in messages:
        author = msg.get("sender", "unknown")

        content_list = msg.get("content", {})
        for content in content_list:
            if content.get("type", "") == "text":
                date_str = content.get("start_timestamp", "")
                date = _iso_to_datev(date_str)

                para = entry.append_paragraph(author)
                para.set_alignment(VT.PA.RIGHT if author == "human" else VT.PA.LEFT)
                para.set_hrule(True)
                para.set_heading_level(VT.PHS.LARGE)

                para = entry.append_paragraph(f"{Date.format_str(date, "F, T")}")
                para.set_alignment(VT.PA.RIGHT if author == "human" else VT.PA.LEFT)

                text = content.get("text", "")
                mdp.import_markdown(entry, text)

                entry.append_paragraph("") # add an empty para for stronger separation





# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import docx
from docx.shared import Inches, RGBColor, Pt
from docx.oxml import OxmlElement
from docx.oxml.ns import qn
from docx.opc.constants import RELATIONSHIP_TYPE as RT
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.enum.table import WD_TABLE_ALIGNMENT, WD_CELL_VERTICAL_ALIGNMENT
from urllib.parse import urlparse, unquote

NAME="Export Docx"
VERSION="0.1"
CATEGORY=SCtg.EXPORTER

# get the default extension:
def get_ext():
    return "docx"

# main func:
def run(path, entries):
    doc = docx.Document()
    disable_page_break = True
    for entry in entries:
        _export_entry_to_docx(doc, entry, disable_page_break)
        disable_page_break = False
    doc.save(path)
    return True


def _export_entry_to_docx(doc, entry, disable_page_break):
    # list styles
    LIST_DEFS = {
        VT.PLS.BULLET:       "bullet",
        VT.PLS.NUMBER:       "decimal",
        VT.PLS.LETTER:       "upperLetter",
        VT.PLS.LETTER_SMALL: "lowerLetter",
        VT.PLS.ROMAN:        "upperRoman",
        VT.PLS.ROMAN_SMALL:  "lowerRoman",
    }
    num_ids = {}
    for k, fmt in LIST_DEFS.items():
        num_ids[k] = _create_numbering(doc, fmt)

    # paragraph loop
    for p in entry:
        text = p.get_text()

        # horizontal rules
        if len(text) == 0 and p.has_hrule():
            _add_horizontal_rule(doc)
            continue

        # images
        if p.is_image(VT.PS.IMAGE_FILE):
            _add_image_paragraph(doc, p)
            continue
        # TODO: add support for charts and tables

        # headings
        heading = p.get_heading_level()
        match heading:
            case VT.PHS.TITLE:
                if not disable_page_break:
                    doc.add_page_break()
                para_w = doc.add_heading(level=1)
            case VT.PHS.LARGE:
                para_w = doc.add_heading(level=2)
            case VT.PHS.MEDIUM:
                para_w = doc.add_heading(level=3)
            case _:
                para_w = doc.add_paragraph()

        if p.get_quot_type() != VT.QT.OFF:
            _set_para_quote(para_w)

        # alignment
        _apply_alignment(para_w, p.get_alignment())

        # list styles
        list_type = p.get_list_type()
        indent = p.get_indent_level()
        match list_type:
            case (VT.PLS.BULLET | VT.PLS.NUMBER | VT.PLS.LETTER | VT.PLS.LETTER_SMALL |
                  VT.PLS.ROMAN | VT.PLS.ROMAN_SMALL):
                _apply_list(para_w, p.get_list_type(), p.get_indent_level(), num_ids)
                indent = 0
            case VT.PLS.TODO:
                para_w.add_run("⏹️ ")
            case VT.PLS.PROGRESSED:
                para_w.add_run("🔄 ")
            case VT.PLS.DONE:
                para_w.add_run("✅ ")
            case VT.PLS.CANCELED:
                para_w.add_run("❌ ")

        # indentation
        if indent > 0:
            para_w.paragraph_format.left_indent = Inches(0.3 * indent)

        # formats & text
        formats = p.get_formats()
        _apply_text_with_formats(para_w, text, formats, list_type, p.get_quot_type() != VT.QT.OFF)


def _apply_text_with_formats(para, text, formats, ps, is_code):
    def _add_hyperlink(paragraph, text, url):
        part = paragraph.part
        r_id = part.relate_to(url, RT.HYPERLINK, is_external=True)

        # create real run
        run = paragraph.add_run(text)

        # now wrap the run in a hyperlink
        hyperlink = OxmlElement("w:hyperlink")
        hyperlink.set(qn("r:id"), r_id)

        hyperlink.append(run._r)

        paragraph._p.append(hyperlink)

        return run


    boundaries = {0, len(text)}
    for f in formats:
        boundaries.add(f.pos_bgn)
        boundaries.add(f.pos_end)
    boundaries = sorted(boundaries)

    for i in range(len(boundaries) - 1):
        bgn, end = boundaries[i], boundaries[i + 1]
        if bgn == end:
            continue

        span = text[bgn:end]
        active_formats = [f for f in formats if f.pos_bgn <= bgn and f.pos_end >= end]

        link = next((f for f in active_formats if f.type == VT.FMT.LINK_URI), None)
        # TODO: add links to LINK_ONTHEFLYs when uri does not start with file://

        # create run
        if is_code:
            run = _add_code_run(para, span)
        elif link:
            run = _add_hyperlink(para, span, link.uri)
        else:
            run = para.add_run(span)

        # hard reset (critical)
        run.font.bold = False
        run.font.italic = False
        run.font.underline = bool(link)
        run.font.strike = False
        run.font.subscript = False
        run.font.superscript = False
        run.font.highlight_color = None

        # styles from list type
        match ps:
            case VT.PLS.TODO:
                run.font.color.rgb = RGBColor(230, 33, 33)
            case VT.PLS.DONE:
                run.font.highlight_color = docx.enum.text.WD_COLOR_INDEX.BRIGHT_GREEN
            case VT.PLS.CANCELED:
                run.font.strike = True

        # apply formats
        for f in active_formats:
            match f.type:
                case VT.FMT.BOLD:
                    run.bold = True
                case VT.FMT.ITALIC:
                    run.italic = True
                case VT.FMT.UNDERLINE:
                    run.underline = True
                case VT.FMT.STRIKETHRU:
                    run.font.strike = True
                case VT.FMT.HIGHLIGHT:
                    run.font.highlight_color = docx.enum.text.WD_COLOR_INDEX.YELLOW
                case VT.FMT.FADED:
                    run.font.color.rgb = RGBColor(128, 128, 128)
                case VT.FMT.SUPERSCRIPT:
                    run.font.superscript = True
                    run.font.subscript = False
                case VT.FMT.SUBSCRIPT:
                    run.font.subscript = True
                    run.font.superscript = False
                case VT.FMT.COMMENT:
                    run.font.color.rgb = RGBColor(128, 128, 128)
                    run.font.size = Pt(8)
                case VT.FMT.CODE_KEYWORD:
                    run.font.color.rgb = RGBColor(255, 0, 0)
                    run.bold = True
                case VT.FMT.CODE_STRING:
                    run.font.color.rgb = RGBColor(64, 224, 64)
                case VT.FMT.CODE_COMMENT:
                    run.font.color.rgb = RGBColor(128, 128, 128)
                case VT.FMT.TAG:
                    run.font.name = "Consolas"
                    run.font.highlight_color = docx.enum.text.WD_COLOR_INDEX.GRAY_25
                case VT.FMT.LINK_URI | VT.FMT.LINK_EVAL | VT.FMT.LINK_ONTHEFLY:
                    run.font.color.rgb = RGBColor(0, 0, 255)

def _add_horizontal_rule(document, thickness=6):
    p = document.add_paragraph()
    p_pr = p._p.get_or_add_pPr()

    p_bdr = OxmlElement("w:pBdr")

    bottom = OxmlElement("w:bottom")
    bottom.set(qn("w:val"), "single")
    bottom.set(qn("w:sz"), str(thickness))   # line thickness
    bottom.set(qn("w:space"), "1")
    bottom.set(qn("w:color"), "auto")

    p_bdr.append(bottom)
    p_pr.append(p_bdr)

def _apply_alignment(para_w, align):
    match align:
        case VT.PA.CENTER:
            para_w.alignment = WD_ALIGN_PARAGRAPH.CENTER
        case VT.PA.RIGHT:
            para_w.alignment = WD_ALIGN_PARAGRAPH.RIGHT
        case _:
            para_w.alignment = WD_ALIGN_PARAGRAPH.LEFT

def _create_numbering(document, num_fmt):
    numbering = document.part.numbering_part.element

    # next available IDs
    abstract_ids = [
        int(n.get(qn("w:abstractNumId")))
        for n in numbering.findall(qn("w:abstractNum"))
    ]
    abstract_id = max(abstract_ids, default=-1) + 1

    num_ids = [
        int(n.get(qn("w:numId")))
        for n in numbering.findall(qn("w:num"))
    ]
    num_id = max(num_ids, default=-1) + 1

    # abstract numbering
    abstract = OxmlElement("w:abstractNum")
    abstract.set(qn("w:abstractNumId"), str(abstract_id))

    for level in range(0, 9):
        lvl = OxmlElement("w:lvl")
        lvl.set(qn("w:ilvl"), str(level))

        # numbering format
        fmt = OxmlElement("w:numFmt")
        fmt.set(qn("w:val"), num_fmt)

        # numbering text (always %1., %2., etc.)
        txt = OxmlElement("w:lvlText")
        txt.set(qn("w:val"), f"%{level + 1}. ")

        # 🔧 critical: avoid Word's default TAB
        suff = OxmlElement("w:suff")
        suff.set(qn("w:val"), "space")

        lvlJc = OxmlElement("w:lvlJc")
        lvlJc.set(qn("w:val"), "right")

        # paragraph properties
        pPr = OxmlElement("w:pPr")
        ind = OxmlElement("w:ind")

        # values in twips
        left = 720 + level * 360   # grows per nesting
        hanging = 360              # keeps marker–text tight

        ind.set(qn("w:left"), str(left))
        ind.set(qn("w:hanging"), str(hanging))

        pPr.append(ind)

        lvl.extend([fmt, txt, suff, lvlJc, pPr])
        abstract.append(lvl)

    numbering.append(abstract)

    # concrete numbering
    num = OxmlElement("w:num")
    num.set(qn("w:numId"), str(num_id))

    abs_id = OxmlElement("w:abstractNumId")
    abs_id.set(qn("w:val"), str(abstract_id))

    num.append(abs_id)
    numbering.append(num)

    return num_id

def _apply_list(para_w, list_type, indent, num_ids):
    if not list_type:
        return  # not a list paragraph

    if list_type not in num_ids:
        return  # unsupported / unknown list type

    pPr = para_w._p.get_or_add_pPr()

    numPr = OxmlElement("w:numPr")

    ilvl = OxmlElement("w:ilvl")
    ilvl.set(qn("w:val"), str(indent))

    numId = OxmlElement("w:numId")
    numId.set(qn("w:val"), str(num_ids[list_type]))

    numPr.append(ilvl)
    numPr.append(numId)
    pPr.append(numPr)

def _set_para_quote(para_w):
    # paragraph shading (light gray)
    pPr = para_w._p.get_or_add_pPr()
    shd = OxmlElement("w:shd")
    shd.set(qn("w:val"), "clear")
    shd.set(qn("w:color"), "auto")
    shd.set(qn("w:fill"), "EDEDED")
    pPr.append(shd)

    # spacing before/after
    para_w.paragraph_format.space_before = Pt(6)
    para_w.paragraph_format.space_after = Pt(6)

def _add_code_run(para_w, text):
    # para_w.style = "Normal"

    run = para_w.add_run(text)

    # monospace
    run.font.name = "Consolas"
    run.font.size = Pt(10)

    # prevent ligatures / smart substitutions
    run.font._element.rPr.append(OxmlElement("w:noProof"))

    return run


def _add_image_paragraph(doc, para):
    path = _uri_to_path(para.get_uri())
    # text = para.get_text().strip()
    align = para.get_alignment()

    if align == VT.PS.ALIGN_C:
        table = doc.add_table(rows=2, cols=1)
        img_cell = table.cell(0, 0)
        txt_cell = table.cell(1, 0)

    elif align == VT.PS.ALIGN_R:
        table = doc.add_table(rows=1, cols=2)
        txt_cell = table.cell(0, 0)
        img_cell = table.cell(0, 1)

    else:  # ALIGN_L (default)
        table = doc.add_table(rows=1, cols=2)
        img_cell = table.cell(0, 0)
        txt_cell = table.cell(0, 1)

    table.alignment = WD_TABLE_ALIGNMENT.CENTER

    _insert_image_into_cell(img_cell, path)
    _insert_text_into_cell(txt_cell, para)

def _insert_image_into_cell(cell, path):
    para_w = cell.paragraphs[0]
    run = para_w.add_run()
    run.add_picture(path, width=Inches(3))
    para_w.alignment = WD_ALIGN_PARAGRAPH.CENTER
    cell.vertical_alignment = WD_CELL_VERTICAL_ALIGNMENT.CENTER

def _insert_text_into_cell(cell, para):
    para_w = cell.paragraphs[0]
    match para.get_alignment():
        case VT.PS.ALIGN_L:
            para_w.alignment = WD_ALIGN_PARAGRAPH.LEFT
        case VT.PS.ALIGN_C:
            para_w.alignment = WD_ALIGN_PARAGRAPH.CENTER
        case VT.PS.ALIGN_R:
            para_w.alignment = WD_ALIGN_PARAGRAPH.RIGHT

    if para.get_text():
        _apply_text_with_formats(para_w, para.get_text(), para.get_formats(),
                                 para.get_list_type(), False)

def _uri_to_path(uri: str) -> str:
    parsed = urlparse(uri)
    return unquote(parsed.path)



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import json
import datetime
import markdown_parsing as mdp

NAME="ChatGPT History Json (Experimental)"
VERSION="0.1"
CATEGORY=SCtg.IMPORTER

_NOW = Date.get_now()

def run(path):
    dm = Diary.get_main()

    with open(path, "r", encoding="utf-8") as f:
        data = json.load(f)

    entry_root = dm.create_entry_rel(None, True, "ChatGPT Import")
    entry_root.append_paragraph(f"Imported on {Date.format_str(_NOW, 'F, T')}")

    # export may contain either:
    # - a list of conversations
    # - or a single conversation
    if isinstance(data, dict):
        conversations = [data]
    else:
        conversations = data

    for conv in conversations:
        title = conv.get("title", "Untitled Conversation")
        if not title: #skip empty conversatons <- TODO CHECK
            continue

        mapping   = conv.get("mapping", {})
        messages  = _get_nodes(mapping)
        date_str  = conv.get("create_time") or ""
        uuid      = conv.get("conversation_id", "")

        _import_entry(dm, entry_root, title, _timestamp_to_datev(date_str), messages, uuid)

    show_message("Import complete!")
    refresh_panels()

    return True

# HELPERS
def _get_nodes(mapping):
    proto_nodes = []

    # collect all real messages:
    for node_id, node in mapping.items():

        msg = node.get("message")

        if msg is None:
            continue

        proto_nodes.append({
            "id": node_id,
            "parent": node.get("parent"),
            "message": msg
        })

    # build quick lookup:
    by_parent = {}

    for node in proto_nodes:
        parent = node["parent"]

        if parent not in by_parent:
            by_parent[parent] = []

        by_parent[parent].append(node)

    # find first actual message:
    current = None

    for node in proto_nodes:
        parent = node["parent"]

        # first message's parent usually points to the artificial root:
        if parent not in [n["id"] for n in proto_nodes]:
            current = node
            break

    if current is None:
        return []

    sequence = []

    visited = set()

    while current:
        node_id = current["id"]

        if node_id in visited:
            print("Cycle detected")
            break

        visited.add(node_id)

        sequence.append(current["message"])

        children = by_parent.get(node_id, [])

        if not children:
            break

        # choose latest child if branching exists:
        best = None
        best_time = -1

        for child in children:
            t = child["message"].get(
                "create_time",
                0
            ) or 0

            if t > best_time:
                best = child
                best_time = t

        current = best

    # fill in the actual nodes to be returned:
    nodes = []
    for message in sequence:
        author = (
            message.get("author", {})
                   .get("role", "unknown")
        )

        content = _extract_text(message)

        if not content.strip():
            continue

        timestamp = message.get("create_time")

        nodes.append({
            "author": author,
            "content": content,
            "date": _timestamp_to_datev(timestamp)
        })

    return nodes

def _extract_text(message):
    content = message.get("content", {})

    parts = content.get("parts", [])

    texts = []

    for p in parts:
        if isinstance(p, str):
            texts.append(p)

    return "\n".join(texts)

def _timestamp_to_datev(timestamp):
    if timestamp:
        try:
            dt = datetime.datetime.fromtimestamp(timestamp)
        except (ValueError, TypeError):
            pass
        else:
            return Date.make(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
    return _NOW


def _import_entry(diary, entry_root, title, date_c, messages, uuid):
    entry = diary.create_entry_rel(entry_root, True, title)
    entry.set_creation_date(date_c)
    para = entry.append_paragraph(uuid)
    para.add_link(f"https://chatgpt.com/c/{uuid}", 0, para.get_size())
    #para.add_format(VT.FMT.FADED)

    for msg in messages:
        author  = msg["author"]
        date    = msg["date"]
        text    = msg["content"]

        para = entry.append_paragraph(author)
        para.set_alignment(VT.PA.RIGHT if author == "user" else VT.PA.LEFT)
        para.set_hrule(True)
        para.set_heading_level(VT.PHS.LARGE)

        para = entry.append_paragraph(f"{Date.format_str(date, "F, T")}")
        para.set_alignment(VT.PA.RIGHT if author == "user" else VT.PA.LEFT)

        mdp.import_markdown(entry, text)

        entry.append_paragraph("") # add an empty para for stronger separation




# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import webbrowser

NAME="Open Location in Google Maps"
VERSION="0.1"
CATEGORY=SCtg.PARA_MODIFIER

def run(paras):
    para = paras[0]
    if para.has_location():
        lat = para.get_latitude()
        lon = para.get_longitude()
        webbrowser.open(f"https://www.google.com/maps?q={lat},{lon}")
    else:
        show_message("Selected paragraph does not have a location!")



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import re
import importer_common as ic

VERSION="0.1"

def import_markdown(entry, raw_lines):
    _ORDERED_RE = re.compile(r"^(\d+)\.\s+")
    in_code_block = False
    code_lang = None

    LIST_MAP = {
        "- ": VT.PLS.BULLET,
        "+ ": VT.PLS.BULLET,
        "* ": VT.PLS.BULLET,
        "A. ": VT.PLS.LETTER,
        "a. ": VT.PLS.LETTER_SMALL,
        "I. ": VT.PLS.ROMAN,
        "i. ": VT.PLS.ROMAN_SMALL,
    }

    for raw_line in raw_lines.splitlines():
        line = raw_line.strip()
        if not line:
            continue

        # indentation:
        indent_spaces = len(raw_line) - len(raw_line.lstrip(" "))
        indent_level = indent_spaces // 2

        # indented code block:
        if raw_line.startswith("    ") and not in_code_block:
            p = entry.append_paragraph(line)
            p.set_quot_type(VT.QT.GENERIC)
            p.set_indent_level(1)
            continue
        # fenced code block:
        elif line.startswith("```"):
            if not in_code_block:
                code_lang = ic.get_code_from_str(line[3:])
                in_code_block = True
            else:
                in_code_block = False
                code_lang = None
            continue

        if in_code_block:
            p = entry.append_paragraph(raw_line)
            p.set_quot_type(code_lang)
            p.set_indent_level(1)
            continue

        # block quote:
        quote_depth, line = _consume_block_quote_prefix(line)
        is_quote = quote_depth > 0
        if is_quote:
            indent_level += (quote_depth-1)

        # horizontal rule
        if _is_horizontal_rule(line):
            p = entry.append_paragraph("")
            p.set_hrule(True)
            #p.set_indent_level(indent_level) TODO: not supported in LoG
            continue

        # heading
        h_level = len(line) - len(line.lstrip("#"))
        if h_level > 0:
            line = line[h_level:].lstrip()

        # to-do
        ps = None
        if line.startswith("- ["):
            mark = line[3]
            ps = {
                " ": VT.PLS.TODO,
                "-": VT.PLS.PROGRESSED,
                "x": VT.PLS.DONE,
                "~": VT.PLS.CANCELED,
            }.get(mark)
            line = line[6:].lstrip()

        # ordered list
        m = _ORDERED_RE.match(line)
        if m:
            ps = VT.PLS.NUMBER
            line = line[m.end():]

        # other lists
        if ps is None:
            for k, v in LIST_MAP.items():
                if line.startswith(k):
                    ps = v
                    line = line[len(k):]
                    break

        if ps:
            indent_level += 1

        # paragraph
        p = entry.append_paragraph("")
        p.set_indent_level(indent_level)

        if is_quote:
            p.set_quote(True)
        if ps:
            p.set_list_type(ps)

        match h_level:
            case 1:
                p.set_heading_level(VT.PHS.LARGE)
            case 2:
                p.set_heading_level(VT.PHS.MEDIUM)
            # case 3:
            #     p.set_heading_level(VT.PHS.SMALL) TODO...
            case _:
                p.add_format(VT.FMT.BOLD)

        _parse_inline_formats(p, line)


# inline parsing (overlap-safe, code-isolated)
def _parse_inline_formats(p, text):
    i = 0
    out = []
    stack = []

    def open_fmt(fmt):
        stack.append((fmt, len(out)))

    def close_fmt(fmt):
        for j in range(len(stack) - 1, -1, -1):
            if stack[j][0] == fmt:
                _, start = stack.pop(j)
                p.add_format(fmt, start, len(out))
                return

    while i < len(text):
        # link
        if text[i] == "[":
            label_start = len(out)
            i += 1
            while i < len(text) and text[i] != "]":
                out.append(text[i])
                i += 1
            label_end = len(out)

            if i + 1 < len(text) and text[i + 1] == "(":
                i += 2
                uri = ""
                while i < len(text) and text[i] != ")":
                    uri += text[i]
                    i += 1
                p.add_format(VT.FMT.LINK_URI, label_start, label_end).uri = uri
                i += 1
            continue

        # underline
        if text.startswith("<u>", i):
            open_fmt(VT.FMT.UNDERLINE)
            i += 3
            continue
        if text.startswith("</u>", i):
            close_fmt(VT.FMT.UNDERLINE)
            i += 4
            continue

        # bold / italic / strikethru / highlight
        TOKENS = [
            ("**", VT.FMT.BOLD),
            ("__", VT.FMT.BOLD),
            ("~~", VT.FMT.STRIKETHRU),
            ("==", VT.FMT.HIGHLIGHT),
            ("`", VT.FMT.ITALIC), # no actual support for inline code
            ("*", VT.FMT.ITALIC),
            ("_", VT.FMT.ITALIC),
        ]

        matched = False
        for tok, fmt in TOKENS:
            if not text.startswith(tok, i):
                continue

            if stack and stack[-1][0] == fmt:
                if tok.startswith("_"):
                    if ic.is_boundary_alnum_after(text, i, len(tok)):
                        continue
                close_fmt(fmt)
            else:
                if tok.startswith("_"):
                    if ic.is_boundary_alnum_before(text, i):
                        continue
                open_fmt(fmt)

            i += len(tok)
            matched = True
            break

        if matched:
            continue

        out.append(text[i])
        i += 1

    p.set_text("".join(out))

def _is_horizontal_rule(line: str) -> bool:
    s = line.strip()
    if len(s) < 3:
        return False
    return all(c == s[0] for c in s) and s[0] in "-*_"

def _consume_block_quote_prefix(line: str):
    # returns (depth, rest_of_line)
    # supports: >text, > text, >    > text, >> text, etc.
    i = 0
    depth = 0
    n = len(line)

    while i < n:
        if line[i] == ">":
            depth += 1
            i += 1
            # skip optional spaces after each '>'
            while i < n and line[i] == " ":
                i += 1
        else:
            break

    return depth, line[i:]



"""
Paragraph Heading Sizes
"""
from __future__ import annotations
__all__: list[str] = ['LARGE', 'MEDIUM', 'NORMAL_TEXT', 'TITLE']
LARGE: int = 261
MEDIUM: int = 260
NORMAL_TEXT: int = 0
TITLE: int = 263



#!/usr/bin/env python3

import os
import subprocess
import sys

datadir = sys.argv[1]
install_prefix = os.environ['MESON_INSTALL_PREFIX']
icondir = os.path.join(install_prefix, datadir, 'icons', 'hicolor')
mimedatabasedir = os.path.join(install_prefix, datadir, 'mime')


if not os.environ.get('DESTDIR'):
  print('Update icon cache...')
  subprocess.call(['gtk-update-icon-cache', '-f', '-t', icondir])
  print('Update MIME database...')
  subprocess.call(['update-mime-database', mimedatabasedir])



"""
Value types
"""
from __future__ import annotations
from . import ES
from . import ETS
from . import FMT
from . import PA
from . import PHS
from . import PLS
from . import PS
from . import QT
__all__: list[str] = ['ES', 'ETS', 'FMT', 'PA', 'PHS', 'PLS', 'PS', 'QT']



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import re

NAME="Format Dates to YYYY*MM*DD"
VERSION="0.1"
CATEGORY=SCtg.ENTRY_MODIFIER

# variables to be shared with the second call (run2):
name_opt = "Day/Month order"
option_default = "DAY/MONTH/YEAR"
entries = []

def run(entries_arg):
    global entries
    entries = entries_arg
    fields = [ { "type": FT.OPTION,
                 "name": name_opt, "opts": [option_default, "MONTH/DAY/YEAR"] } ]
    get_user_input( "Please specify Day/Month order in the existing dates", NAME, "run2", fields)

def run2(user_info : Dict[str, Any]):
    # captures D or DD, separator, M or MM, separator, YYYY:
    date_re = re.compile(r"\b(\d{1,2})[./](\d{1,2})[./](\d{4})\b")
    # day/month priority based on user input:
    do, mo = (1, 2) if user_info[name_opt] == option_default else (2, 1)

    # nested func to do the actual work:
    def format_dates(entry):
        for para in entry:
            text = para.get_text()
            for match in reversed(list(date_re.finditer(text))):
                d, m, year = match.group(do), match.group(mo), match.group(3)
                dt = Date.make(int(year), int(m), int(d))
                # if the preferred day/month order did not work, try the other one as a last resort:
                if not Date.is_valid(dt):
                    dt = Date.make(int(year), int(d), int(m))
                    if not Date.is_valid(dt):
                        continue
                new_text = Date.format_str(dt)
                start, end = match.start(), match.end()
                para.replace_text(start, end - start, new_text)

    for entry in entries:
        format_dates(entry)

    refresh_panels()



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import os
from html import escape

NAME="Export to HTML"
VERSION="0.1"
CATEGORY=SCtg.EXPORTER

# get the default extension:
def get_ext():
    return "html"

# main func:
def run(path, entries):
    source = f"""<!DOCTYPE html>
<html>
<head>
<meta charset="utf-8">
<title>{os.path.basename(path)}</title>
<style>
{CSS}
</style>
</head>
<body>
<article class="entry">
"""
    for entry in entries:
        source += _export_entry_to_html(entry)
    source += """
</article>
</body>
</html>
"""

    with open(path, "w", encoding="utf-8") as f:
        f.write(source)

    return True


# main exporter
def _export_entry_to_html(entry):
    source = []
    list_stack = []  # [(ps_type, indent, tag)]
    ti = 0 # tag indent
    quot_type_last = VT.QT.OFF

    def _ins_tag_ind(offset_bgn, offset_end):
        nonlocal ti
        ti_bgn = ti + offset_bgn
        ti += offset_end
        return "  " * ti_bgn

    def _append_html(new_html, offset_bgn=1, offset_end=1):
        source.append(f"{_ins_tag_ind(offset_bgn, offset_end)}{new_html}")

    def _close_lists(to_indent=0):
        while list_stack and list_stack[-1][1] >= to_indent:
            tag = list_stack.pop()[2]
            _append_html(f"</{tag}>", 0, -1)

    def _close_open_tags(qt): # TODO: combine with the above one
        nonlocal quot_type_last

        # quotation
        if quot_type_last != qt:
            if quot_type_last == VT.QT.LITERARY:
                _append_html("</div>", 0, -1)
            # code
            elif quot_type_last != VT.QT.OFF:
                _append_html("</code></pre>", 0, -1)
                _append_html("</div>", 0, -1)
            quot_type_last = VT.QT.OFF


    for p in entry:
        text = p.get_text()
        indent = p.get_indent_level()
        pls = p.get_list_type()
        formats = p.get_formats()
        quot_type = p.get_quot_type()

        _close_open_tags(quot_type)

        # horizontal rule
        if p.has_hrule() and p.is_empty_completely():
            _append_html("<hr>", 1, 0)
            continue
        # TODO: find a solution for hrules with text

        # images
        if p.is_image( VT.PS.IMAGE_FILE):
            IMG_ALIGN_CLASS = {
                VT.PA.LEFT: "img-left",    # text right
                VT.PA.RIGHT: "img-right",   # text left
                VT.PA.CENTER: "img-center",  # text below
            }

            uri = p.get_uri()
            align = p.get_alignment()
            cls = IMG_ALIGN_CLASS.get(align, "img-bottom")

            _append_html(f'<div class="image-block {cls}">')
            _append_html(f'<img src="{escape(uri)}">', 1, 0)

            if text:
                _append_html(f'<div class="image-text">{_emit_inline(text, formats)}</div>', 1, 0)

            _append_html('</div>', 0, -1)
            continue
        # TODO: add support for charts and tables

        # code
        if p.is_code():
            if quot_type_last != quot_type:
                _append_html('<div class="quote-frame">')
                _append_html(f'<div class="code-header">{p.get_code_lang()}</div>', 1, 0)
                _append_html('<pre><code>')
                quot_type_last = quot_type
            source.append(f"{_emit_inline(text, formats)}") # raw append
            continue

        # headings
        heading = _get_heading_level(p)
        if heading != 0:
            _close_lists(0) # FIXME: in Lifeograph a header can as well be a list item
            _append_html(f"<h{heading}{_get_classes(p)}>{_emit_inline(text, formats)}</h{heading}>",
                        1, 0)
            continue

        # quotation
        if p.is_quote() and quot_type_last != VT.QT.LITERARY:
            _append_html('<div class="quote-frame">')
            quot_type_last = quot_type

        # lists
        tag, cls = _list_tag(pls)
        if tag:
            while list_stack and (list_stack[-1][1] > indent or
                                  (list_stack[-1][1] == indent and list_stack[-1][3] != cls)):
                _append_html(f"</{list_stack.pop()[2]}>", 0, -1)

            while not list_stack or list_stack[-1][1] < indent:
                _append_html(f"<{tag}{cls}>")
                list_stack.append((pls, indent, tag, cls))

            _append_html(f"<li>{_emit_inline(text, formats)}</li>", 1, 0)
            continue

        # normal paragraph / todo
        _close_lists(0)

        if pls in (VT.PLS.TODO, VT.PLS.PROGRESSED, VT.PLS.DONE, VT.PLS.CANCELED):
            state = {
                VT.PLS.TODO: "todo",
                VT.PLS.PROGRESSED: "progressed",
                VT.PLS.DONE: "done",
                VT.PLS.CANCELED: "canceled",
            }[pls]

            _append_html(
                f'<p class="todo indent-{indent}" data-state="{state}">'
                f'<span class="todo-marker"></span>'
                f'{_emit_inline(text, formats)}</p>', 1, 0)
        else:
            _append_html( f'<p{_get_classes(p)}>{_emit_inline(text, formats)}</p>', 1, 0)

    _close_open_tags(VT.QT.OFF)
    _close_lists(0)

    return "\n".join(source)

# inline formatting (overlap-safe)
def _emit_inline(text, formats):
    boundaries = {0, len(text)}
    for f in formats:
        boundaries.add(f.pos_bgn)
        boundaries.add(f.pos_end)

    points = sorted(boundaries)
    out = []

    for i in range(len(points) - 1):
        a, b = points[i], points[i + 1]
        if a == b:
            continue

        chunk = escape(text[a:b])
        active = [f for f in formats if f.pos_bgn <= a and f.pos_end >= b]

        # link outermost
        link = next((f for f in active if f.type == VT.FMT.LINK_URI), None)
        if link:
            chunk = f'<a href="{escape(link.uri)}">{chunk}</a>'

        for f in active:
            match f.type:
                case VT.FMT.BOLD:
                    chunk = f"<strong>{chunk}</strong>"
                case VT.FMT.ITALIC:
                    chunk = f"<em>{chunk}</em>"
                case VT.FMT.UNDERLINE:
                    chunk = f"<u>{chunk}</u>"
                case VT.FMT.STRIKETHRU:
                    chunk = f"<s>{chunk}</s>"
                case VT.FMT.HIGHLIGHT:
                    chunk = f"<mark>{chunk}</mark>"
                case VT.FMT.FADED:
                    chunk = f'<span class="faded">{chunk}</span>'
                case VT.FMT.SUPERSCRIPT:
                    chunk = f'<sup>{chunk}</sup>'
                case VT.FMT.SUBSCRIPT:
                    chunk = f'<sub>{chunk}</sub>'
                case VT.FMT.COMMENT:
                    chunk = f'<span class="comment">{chunk}</span>'
                case VT.FMT.CODE_KEYWORD:
                    chunk = f'<span class="keyword">{chunk}</span>'
                case VT.FMT.CODE_STRING:
                    chunk = f'<span class="code_string">{chunk}</span>'
                case VT.FMT.CODE_COMMENT:
                    chunk = f'<span class="comment">{chunk}</span>'
                case VT.FMT.TAG:
                    chunk = f'<span class="lifeo_tag">{chunk}</span>'

        out.append(chunk)

    return "".join(out)

# list helpers
def _list_tag(pls):
    if pls == VT.PLS.BULLET:
        return "ul", ""
    if pls == VT.PLS.NUMBER:
        return "ol", ' type="1"'
    if pls == VT.PLS.LETTER:
        return "ol", ' type="A"'
    if pls == VT.PLS.LETTER_SMALL:
        return "ol", ' type="a"'
    if pls == VT.PLS.ROMAN:
        return "ol", ' type="I"'
    if pls == VT.PLS.ROMAN_SMALL:
        return "ol", ' type="i"'
    return None, ""

# css classes for paragraph
def _get_classes(p):
    ind = p.get_indent_level()
    ind_str = f' indent-{ind}' if ind > 0 else ""
    align_str = ""
    match p.get_alignment():
        case VT.PA.CENTER:
            align_str = " align-c"
        case VT.PA.RIGHT:
            align_str = " align-r"
    return f' class="{ind_str}{align_str}"' if ind_str or align_str else ""

def _get_heading_level(p):
    match p.get_heading_level():
        case VT.PHS.TITLE:  return 1
        case VT.PHS.LARGE:  return 2
        case VT.PHS.MEDIUM: return 3
        case _: return 0

CSS = """
body {
  font-family: system-ui, sans-serif;
  line-height: 1.2;
}

.entry {
  max-width: 70ch;
  margin: 2rem auto;
}

.indent-1 { margin-left: 2em; }
.indent-2 { margin-left: 4em; }
.indent-3 { margin-left: 6em; }
.indent-4 { margin-left: 8em; }

.align-l    { text-align: left; }
.align-c    { text-align: center; }
.align-r    { text-align: right; }

.todo {
  display: flex;
  gap: .5em;
}

.todo-marker::before {
  content: "☐";
}

.todo[data-state="todo"] {
  color: red;
}

.todo[data-state="done"] .todo-marker::before {
  content: "✔";
}

.todo[data-state="progressed"] .todo-marker::before {
  content: "◐";
}

.todo[data-state="canceled"] {
  opacity: .6;
  text-decoration: line-through;
}

.faded {
  color: rgba(0, 0, 0, 0.45);
}

.comment {
  color: rgba(0, 0, 0, 0.5);
  font-size: 0.8em;
}

.lifeo_tag {
  font-family: ui-monospace, monospace;
  border: 1px solid #ccc;
  border-radius: 2px;
  background: #f0e0e0;
}

code {
  font-family: ui-monospace, monospace;
  font-size: 0.9em;
}

.quote-frame {
  border: 1px solid #ccc;
  border-radius: 6px;
  margin: 1em 0;
  background: #f8f8f8;
}

.code-header {
  font-size: 0.9em;
  padding: 4px 8px;
  background: #eaeaea;
  border-bottom: 1px solid #ccc;
  font-family: monospace;
}

.quote-frame pre {
  margin: 0;
  padding: 8px;
  overflow-x: auto;
}

.keyword {
  color: rgb(255, 0, 0);
  font-weight: bold;
}

.code_string {
  color: rgb(0, 225, 0);
}

.image-block {
  display: flex;
  gap: 12px;
  margin: 1em 0;
  align-items: flex-start;
}

.image-block img {
  max-width: 40%;
  height: auto;
}

/* text to the right of image */
.image-block.img-left {
  flex-direction: row;
}

/* text to the left of image */
.image-block.img-right {
  flex-direction: row-reverse;
}

/* text below image */
.image-block.img-center {
  flex-direction: column;
  align-items: center;
}

hr {
  border: none;
  border-top: 1px solid #ccc;
  margin: 1.5em 0;
}
"""



"""
Rich Text Formats
"""
from __future__ import annotations
__all__: list[str] = ['BOLD', 'CODE_COMMENT', 'CODE_KEYWORD', 'CODE_STRING', 'COMMENT', 'DATE', 'FADED', 'HIGHLIGHT', 'ITALIC', 'LINK_EVAL', 'LINK_ID', 'LINK_ONTHEFLY', 'LINK_URI', 'STRIKETHRU', 'SUBSCRIPT', 'SUPERSCRIPT', 'TAG', 'TAG_VALUE', 'TIME', 'UNDERLINE']
BOLD: int = 0
CODE_COMMENT: int = 65618
CODE_KEYWORD: int = 65617
CODE_STRING: int = 65619
COMMENT: int = 65551
DATE: int = 17104915
FADED: int = 5
HIGHLIGHT: int = 2
ITALIC: int = 1
LINK_EVAL: int = 786442
LINK_ID: int = 9175051
LINK_ONTHEFLY: int = 327692
LINK_URI: int = 786441
STRIKETHRU: int = 3
SUBSCRIPT: int = 4194310
SUPERSCRIPT: int = 4194311
TAG: int = 8650760
TAG_VALUE: int = 16842766
TIME: int = 16842768
UNDERLINE: int = 4



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import os
import re
from urllib.parse import urlparse, unquote

NAME="Export Markdown"
VERSION="0.1"
CATEGORY=SCtg.EXPORTER

# get the default extension:
def get_ext():
    return "md"

# main func:
def run(path, entries):
    base_dir = os.path.dirname(path) or "."
    base_name = os.path.basename(path)

    has_md = base_name.lower().endswith(".md")
    stem = base_name[:-3] if has_md else base_name

    for entry in entries:
        md = _export_entry_to_markdown(entry)

        filename = (f"{stem}__{_sanitize_filename(entry.get_name())}.md" if len(entries)>1
                    else f"{stem}.md")

        out_path = os.path.join(base_dir, filename)

        with open(out_path, "w", encoding="utf-8") as f:
            f.write(md)

    return True


def _sanitize_filename(name: str) -> str:
    # remove forbidden characters and collapse whitespace
    name = re.sub(r'[\\/:*?"<>|]', "", name)
    name = re.sub(r"\s+", "_", name.strip())
    return name or "entry"

def _uri_to_path(uri: str) -> str:
    parsed = urlparse(uri)
    return unquote(parsed.path)

def _export_entry_to_markdown(entry):
    out = []
    last_para_was_code = False

    for p in entry:
        text = p.get_text()

        # blockquote
        blockquote = ""
        if p.is_quote():
            blockquote = "> " * p.get_indent_level()

        #image
        if p.is_image(VT.PS.IMAGE_FILE):
            path = _uri_to_path(p.get_uri())
            out.append(f'![image]({path})')

        # indentation (only valid for list tems in MD)
        indent = ""
        if p.get_list_type() != VT.PLS.PLAIN:
            indent = "  " * p.get_indent_level()

        # code lang
        lang_str = p.get_code_lang()

        if p.is_code():
            if not last_para_was_code:
                out.append("")
                out.append(f"``` {lang_str}")
                last_para_was_code = True
        elif last_para_was_code:
            out.append("```")
            out.append("")
            last_para_was_code = False

        # heading level
        h = 0
        heading = ""
        match p.get_heading_level():
            case VT.PHS.TITLE:
                h = 1
            case VT.PHS.LARGE:
                h = 1 if p.has_hrule() else 2 #H2 header with hrule is regarded as H1
            case VT.PHS.MEDIUM:
                h = 3
        if h > 0:
            heading = ("#" * h + " ")

        # list / todo prefix
        list_type = ""
        ps = p.get_list_type()

        TODO_MAP = {
            VT.PLS.TODO: "[ ] ",
            VT.PLS.PROGRESSED: "[-] ",
            VT.PLS.DONE: "[x] ",
            VT.PLS.CANCELED: "[~] ",
        }

        LIST_MAP = {
            VT.PLS.BULLET: "- ",
            VT.PLS.NUMBER: "1. ",
            VT.PLS.LETTER: "a. ",
            VT.PLS.LETTER_SMALL: "A. ",
            VT.PLS.ROMAN: "i. ",
            VT.PLS.ROMAN_SMALL: "I. ",
        }

        if ps in TODO_MAP and h==0:
            list_type = "- " + TODO_MAP[ps]
        elif ps in LIST_MAP:
            list_type = LIST_MAP[ps]

        # inline formatting
        formatted = _apply_inline_formats(text, p.get_formats())

        if not (p.is_empty_completely() and p.has_hrule()):
            out.append(f"{blockquote}{indent}{list_type}{heading}{formatted}")

        if p.has_hrule() and p.get_heading_level() != VT.PHS.TITLE:
            out.append("---")

    return "\n".join(out).rstrip() + "\n"


def _apply_inline_formats(text, formats):
    events = []

    for f in formats:
        start, end, fmt = f.pos_bgn, f.pos_end, f.type

        if fmt == VT.FMT.BOLD:
            events.append((start, "**", True))
            events.append((end, "**", False))
        elif fmt == VT.FMT.ITALIC:
            events.append((start, "*", True))
            events.append((end, "*", False))
        elif fmt == VT.FMT.UNDERLINE:
            events.append((start, "<u>", True))
            events.append((end, "</u>", False))
        elif fmt == VT.FMT.STRIKETHRU:
            events.append((start, "~~", True))
            events.append((end, "~~", False))
        elif fmt == VT.FMT.HIGHLIGHT:
            events.append((start, "==", True))
            events.append((end, "==", False))
        elif fmt == VT.FMT.LINK_URI:
            events.append((start, "[", True))
            events.append((end, f"]({f.uri})", False))

    # openings first, closings last at same index
    events.sort(key=lambda e: (e[0], not e[2]))

    result = []
    pos = 0

    for idx, token, opening in events:
        if idx > pos:
            result.append(text[pos:idx])
            pos = idx
        result.append(token)

    result.append(text[pos:])
    return "".join(result)



"""
Paragraph List Styles
"""
from __future__ import annotations
__all__: list[str] = ['BULLET', 'CANCELED', 'DONE', 'LETTER', 'LETTER_SMALL', 'NUMBER', 'PLAIN', 'PROGRESSED', 'ROMAN', 'ROMAN_SMALL', 'TODO']
BULLET: int = 592
CANCELED: int = 1600
DONE: int = 1584
LETTER: int = 2672
LETTER_SMALL: int = 2688
NUMBER: int = 2656
PLAIN: int = 0
PROGRESSED: int = 1568
ROMAN: int = 2704
ROMAN_SMALL: int = 2720
TODO: int = 1552



"""
Date utilities
"""
from __future__ import annotations
import typing
__all__: list[str] = ['format_str', 'get_day', 'get_month', 'get_now', 'get_today', 'get_year', 'is_valid', 'make']
@typing.overload
def format_str(date: typing.SupportsInt | typing.SupportsIndex) -> str:
    """
    Converts a date into a string in the default format
    Parameters:
    date: The date to be converted
    returns: The date string
    """
@typing.overload
def format_str(date: typing.SupportsInt | typing.SupportsIndex, format: str) -> str:
    """
    Converts a date into a string in a custom format.
    Format is a string where special chars are replaced with date fileds as below:
    M: month number
    n: month name
    D: day
    d: week day
    W: week no
    h: hour
    m: minute
    s: second
    F: date in default format
    T: time in default format
    Parameters:
    date: The date to be converted
    format: The format string
    returns: The date string
    """
def get_day(arg0: typing.SupportsInt | typing.SupportsIndex) -> int:
    ...
def get_month(arg0: typing.SupportsInt | typing.SupportsIndex) -> int:
    ...
def get_now() -> int:
    ...
def get_today() -> int:
    ...
def get_year(arg0: typing.SupportsInt | typing.SupportsIndex) -> int:
    ...
def is_valid(arg0: typing.SupportsInt | typing.SupportsIndex) -> bool:
    ...
@typing.overload
def make(year: typing.SupportsInt | typing.SupportsIndex, month: typing.SupportsInt | typing.SupportsIndex, day: typing.SupportsInt | typing.SupportsIndex) -> int:
    ...
@typing.overload
def make(year: typing.SupportsInt | typing.SupportsIndex, month: typing.SupportsInt | typing.SupportsIndex, day: typing.SupportsInt | typing.SupportsIndex, hour: typing.SupportsInt | typing.SupportsIndex, minute: typing.SupportsInt | typing.SupportsIndex, second: typing.SupportsInt | typing.SupportsIndex) -> int:
    ...
@typing.overload
def make(arg0: str) -> int:
    ...



"""
Element Statuses
"""
from __future__ import annotations
__all__: list[str] = ['AUTO', 'CANCELED', 'DONE', 'PROGRESSED', 'TODO']
AUTO: int = 4096
CANCELED: int = 65536
DONE: int = 32768
PROGRESSED: int = 16384
TODO: int = 8192



"""
Entry Title Styles
"""
from __future__ import annotations
__all__: list[str] = ['DATE_AND_NAME', 'INHERIT', 'MILESTONE', 'NAME_AND_DESCRIPT', 'NAME_ONLY', 'NUMBER_AND_NAME']
DATE_AND_NAME: int = 3
INHERIT: int = 5
MILESTONE: int = 4
NAME_AND_DESCRIPT: int = 1
NAME_ONLY: int = 0
NUMBER_AND_NAME: int = 2



"""
Code Languages
"""
from __future__ import annotations
__all__: list[str] = ['BASH', 'CPP', 'FORTRAN', 'GENERIC', 'GO', 'HASKELL', 'HTML', 'JAVA', 'JAVASCRIPT', 'KOTLIN', 'LISP', 'LUA', 'OFF', 'PASCAL', 'PERL', 'PYTHON', 'RUBY', 'RUST', 'SCALA', 'SQL']
BASH: str = 'B'
CPP: str = 'C'
FORTRAN: str = 'F'
GENERIC: str = '*'
GO: str = 'G'
HASKELL: str = 'H'
HTML: str = '<'
JAVA: str = 'J'
JAVASCRIPT: str = 'V'
KOTLIN: str = 'K'
LISP: str = '('
LUA: str = 'U'
OFF: str = '_'
PASCAL: str = 'L'
PERL: str = '$'
PYTHON: str = 'P'
RUBY: str = 'R'
RUST: str = 'T'
SCALA: str = 'S'
SQL: str = 'Q'



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
import webbrowser
import urllib.parse

NAME="Lookup in Wikipedia"
VERSION="0.1"
CATEGORY=SCtg.WORD_MODIFIER

def run(para: Paragraph, pos_bgn, pos_end):
    text = para.get_text()

    # if no selection → expand to word boundaries
    if pos_bgn == pos_end:
        delimiters = " \t\n.,;:!?()[]{}<>\"'`"

        # find start
        start = pos_bgn
        while start > 0 and text[start - 1] not in delimiters:
            start -= 1

        # find end
        end = pos_end
        while end < len(text) and text[end] not in delimiters:
            end += 1

        word = text[start:end]
    else:
        word = text[pos_bgn:pos_end]

    word = word.strip()
    if not word:
        return

    # encode for URL
    encoded = urllib.parse.quote(word)

    # set lookup site url:
    url = f"https://en.wikipedia.org/wiki/{encoded}"

    # alternatives:
    # url = f"https://en.wiktionary.org/wiki/{encoded}"
    # url = f"https://dictionary.cambridge.org/dictionary/english/{encoded}"
    # url = f"https://www.merriam-webster.com/dictionary/{encoded}"
    # url = f"https://www.oxfordlearnersdictionaries.com/definition/english/{encoded}"

    webbrowser.open(url)



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
from pathlib import Path
import xml.etree.ElementTree as ET

NAME="Import KML (Experimental)"
VERSION="0.1"
CATEGORY=SCtg.IMPORTER


def run(path):
    _import_kml_file(path)

    refresh_panels()

    return True


# XML HELPERS (namespace-agnostic) =================================================================
# KML files declare namespaces like xmlns="http://www.opengis.net/kml/2.2" (2.0/2.1/2.3 variants
# also exist), so we match on local tag name rather than building a namespace map.

def _local(tag):
    return tag.split('}', 1)[-1] if '}' in tag else tag

def _find_child(elem, name):
    for child in elem:
        if _local(child.tag) == name:
            return child
    return None

def _get_text(elem, name):
    child = _find_child(elem, name)
    if child is not None and child.text:
        text = child.text.strip()
        return text if text else None
    return None

def _parse_coords(text):
    # KML coordinate tuples are "lon,lat[,alt]", whitespace-separated for multi-point geometries
    coords = []
    for tok in text.split():
        parts = tok.split(',')
        if len(parts) >= 2:
            try:
                lon = float(parts[0])
                lat = float(parts[1])
            except ValueError:
                continue
            coords.append((lat, lon))
    return coords


# GEOMETRY IMPORT ==================================================================================
def _add_description(entry, description):
    if description:
        d = entry.append_paragraph(description)
        d.set_indent_level(1)

def _import_point(entry, point_elem, name, description):
    coords_text = _get_text(point_elem, 'coordinates')
    if not coords_text:
        return
    coords = _parse_coords(coords_text)
    if not coords:
        return

    lat, lon = coords[0]
    p = entry.append_paragraph(name if name else "")
    p.set_location(lat, lon)

    _add_description(entry, description)

def _import_path(entry, coords_elem_parent, name, description):
    # used for LineString and Polygon rings. Consecutive paragraphs at the same indent level
    # with locations set are rendered as a path on the map, so this covers both.
    coords_text = _get_text(coords_elem_parent, 'coordinates')
    if not coords_text:
        return
    coords = _parse_coords(coords_text)
    if not coords:
        return

    first = True
    for lat, lon in coords:
        p = entry.append_paragraph(name if first and name else "")
        p.set_location(lat, lon)
        first = False

    _add_description(entry, description)

def _import_polygon(entry, polygon_elem, name, description):
    outer = _find_child(polygon_elem, 'outerBoundaryIs')
    if outer is None:
        return
    ring = _find_child(outer, 'LinearRing')
    if ring is None:
        return

    _import_path(entry, ring, name, description)
    # TODO: inner boundaries (holes) are not represented (no equivalent concept in Lifeograph paths)

def _import_placemark(entry, placemark):
    name = _get_text(placemark, 'name')
    description = _get_text(placemark, 'description')

    point = _find_child(placemark, 'Point')
    linestring = _find_child(placemark, 'LineString')
    polygon = _find_child(placemark, 'Polygon')
    multigeom = _find_child(placemark, 'MultiGeometry')

    if point is not None:
        _import_point(entry, point, name, description)
    elif linestring is not None:
        _import_path(entry, linestring, name, description)
    elif polygon is not None:
        _import_polygon(entry, polygon, name, description)
    elif multigeom is not None:
        # Import each sub-geometry; only the first gets the shared name/description label
        for geom in multigeom:
            local = _local(geom.tag)
            if local == 'Point':
                _import_point(entry, geom, name, description)
            elif local == 'LineString':
                _import_path(entry, geom, name, description)
            elif local == 'Polygon':
                _import_polygon(entry, geom, name, description)
            else:
                continue
            name, description = None, None
    elif name or description:
        # No geometry at all, still record the label/description as plain paragraphs
        entry.append_paragraph(name if name else "")
        _add_description(entry, description)


# CONTAINER TRAVERSAL ==============================================================================
def _process_container(container, title):
    """Placemarks that sit directly under a Document (or the kml root, or a nested <Document>)
    without belonging to any Folder are collected into one entry named after the container
    (or the file title). Each Folder becomes its own separate entry."""
    loose_entry = None

    def _get_loose_entry():
        nonlocal loose_entry
        if loose_entry is None:
            name = _get_text(container, 'name') or title
            loose_entry = Diary.get_main().create_entry(VT.ETS.NAME_ONLY, name)
        return loose_entry

    for child in container:
        local = _local(child.tag)
        if local == 'Placemark':
            _import_placemark(_get_loose_entry(), child)
        elif local == 'Folder':
            _process_folder(child)
        elif local == 'Document':
            _process_container(child, title)

def _process_folder(folder):
    name = _get_text(folder, 'name') or "Untitled Folder"
    entry = Diary.get_main().create_entry(VT.ETS.NAME_ONLY, name)

    for child in folder:
        local = _local(child.tag)
        if local == 'Placemark':
            _import_placemark(entry, child)
        elif local == 'Folder':
            _process_folder(child)  # nested folder -> its own separate entry, not nested in `entry`

def _import_kml_file(filepath):
    tree = ET.parse(filepath)
    root = tree.getroot()
    title = Path(filepath).stem

    document = _find_child(root, 'Document')
    _process_container(document if document is not None else root, title)



"""
Paragraph Styles
"""
from __future__ import annotations
__all__: list[str] = ['IMAGE_CHART', 'IMAGE_FILE', 'IMAGE_TABLE']
IMAGE_CHART: int = 1179648
IMAGE_FILE: int = 1114112
IMAGE_TABLE: int = 1245184



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
from pathlib import Path
import markdown_parsing as mdp

NAME="Import Markdown"
VERSION="0.2"
CATEGORY=SCtg.IMPORTER

def run(path):
    entry = Diary.get_main().create_entry(VT.ETS.NAME_ONLY, Path(path).stem)

    with open(path, encoding="utf-8") as f:
        raw_lines = f.read()

    mdp.import_markdown(entry, raw_lines)

    refresh_panels()

    return True




"""
LoG module documentation
"""
from __future__ import annotations
import collections.abc
import typing
from . import Date
from . import VT
__all__: list[str] = ['Date', 'Diary', 'DiaryElemTag', 'DiaryElement', 'Entry', 'FT', 'HiddenFormat', 'Named', 'ParaInhClass', 'Paragraph', 'ParserBackGround', 'SCtg', 'TableColumn', 'TableLine', 'VT', 'get_user_input', 'refresh_panels', 'show_element', 'show_message', 'update_for_new_entry']
class Diary(DiaryElement):
    @staticmethod
    def get_main() -> Diary:
        """
        Returns the currently opened main Diary
        """
    def create_entry(self, title_style: typing.SupportsInt | typing.SupportsIndex, text: str = '') -> Entry:
        """
        Creates an empty Entry and adds it to the end of hierarchy
        Parameters:
        title_style: Title style (VT.ETS) to be used in the created Entry
        text: Contents for the new Entry (optional)
        returns: the newly created Entry
        """
    def create_entry_child(self, entry_parent: Entry, text: str = '') -> Entry:
        """
        Creates an Entry with given text as the last child of entry_parent
        Parameters:
        entry_parent: Parent Entry
        text: Contents for the new Entry (optional)
        returns: The newly created Entry
        """
    def create_entry_rel(self, entry_rel: Entry, is_parent: bool, text: str = '') -> Entry:
        """
        Creates an Entry with given text and places it in the specified position
        Parameters:
        entry_rel: Related Entry, a parent or sibling. If it is None, Entry will be added as a top level
        is_parent: If True entry_rel is a parent else a sibling. If True, Entry is added as the first child to the entry_rel. If entry_rel is None, determines if entry will be added to the beginning (False) or end (True)
        text: Contents for the new Entry (optional)
        returns: The newly created Entry
        """
    def get_entry_by_id(self, id: typing.SupportsInt | typing.SupportsIndex) -> Entry:
        """
        Returns the Entry with the given id if any
        """
    def get_entry_by_name(self, name: str) -> Entry:
        """
        Returns the first Entry with the given name
        Please note that there may be multiple entries sharing a name
        """
    def get_entry_by_name_fuzzy(self, name: str) -> Entry:
        """
        Returns the Entry with the name most closely matching the argument
        """
    def get_paragraph_by_id(self, id: typing.SupportsInt | typing.SupportsIndex) -> Paragraph:
        """
        Returns the Paragraph with the given id if any
        """
    def get_rel_folder_name(self) -> str:
        """
        Returns the name of the relative folder for the current entry which takes the form [DIARYNAME] (sanitized)
        """
    def get_rel_folder_path(self) -> str:
        """
        Returns path to the relative folder for the current entry which is a subfolder in the Diary's folder in the form [DIARYNAME]
        """
    def get_size(self) -> int:
        """
        Returns the number of entries in the Diary
        """
    def get_tag_by_id(self, id: typing.SupportsInt | typing.SupportsIndex) -> DiaryElemTag:
        """
        Returns the DiaryElemTag with the given id if any
        """
class DiaryElemTag(DiaryElement):
    """
    Base class for Paragraphs and Entries.
    
    Diary elements are objects stored inside a diary.
    They have an ID and occupy space.
    """
    def get_ancestry_path(self) -> str:
        ...
    def get_date(self) -> int:
        ...
    def get_date_finish(self) -> int:
        ...
    def get_description(self) -> str:
        ...
    def is_expanded(self) -> bool:
        """
        Returns true if the entry/paragraph is expanded (i.e. shows its children)
        """
    def set_expanded(self, is_expanded: bool) -> None:
        """
        For parent entries/tags, sets the collapsed/expanded state
        Parameters:
        is_expanded: Supply True to expanded or False to collapse
        """
class DiaryElement(Named):
    """
    Base class for all diary elements.
    
    Diary elements are objects stored inside a diary.
    They have an ID and occupy space.
    """
    def get_id(self) -> int:
        ...
    def get_size(self) -> int:
        ...
class Entry(DiaryElemTag):
    def __iter__(self) -> collections.abc.Iterator[Paragraph]:
        ...
    def add_paragraph_before(self, text: str, para_after: Paragraph) -> Paragraph:
        """
        Creates a paragraph and places it before another
        Parameters:
        text: The content of the new paragraph (mustn't contain new line char)
        para_after: The paragraph before which the new para will be inserted
        returns: The newly created paragraph
        """
    def add_paragraphs_after(self, para_first_to_add: Paragraph, para_before: Paragraph, inheritance_classes: ParaInhClass = ...) -> Paragraph:
        """
        Add Paragraph chain starting with para_first_to_add
        """
    def append_paragraph(self, text: str) -> Paragraph:
        """
        Creates a paragraph and adds it to the end of the entry
        Parameters:
        text: The content of the new paragraph (mustn't contain new line char)
        returns: The newly created paragraph
        """
    def beautify(self, para_begin: Paragraph, para_end: Paragraph) -> None:
        """
        Beautify the text in the specified range
        Converts * to bullet, removes excessive empty lines, etc..
        Parameters:
        para_begin: first paragraph to beautify
        para_end: last paragraph to beautify
        """
    def get_completion(self) -> float:
        """
        Returns the completion percentage of the Entry as a float between in [0,1] interval
        """
    def get_completion_str(self) -> str:
        """
        Get Entry completion as a string with both percentage and fraction
        """
    def get_paragraph(self, para_no: typing.SupportsInt | typing.SupportsIndex) -> Paragraph:
        """
        Returns a Paragraph from the Entry by order (first paragraph is 0)
        """
    def get_paragraph_last(self) -> Paragraph:
        """
        Returns the last Paragraph of the Entry
        """
    def get_parent(self) -> Entry:
        """
        Returns the Entry's parent Entry if any or None
        """
    def get_text(self) -> str:
        """
        Returns the Entry's text
        """
    def get_todo_status(self) -> int:
        """
        Returns the todo status of the Entry as a VT.ES.* value
        """
    @typing.overload
    def remove_paragraphs(self, para_begin: Paragraph, para_end: Paragraph = None, add_undo: bool = True) -> None:
        """
        Remove Paragraphs begin, end, and anything between from the Entry
        Only removes a single Paragraph when end is None
        Parameters:
        para_begin: The first Paragraph to remove
        para_end: The last Paragraph to remove
        add_undo: Whether the action will be undoable
        """
    @typing.overload
    def remove_paragraphs(self, para_no_begin: typing.SupportsInt | typing.SupportsIndex, para_no_end: typing.SupportsInt | typing.SupportsIndex = -1, add_undo: bool = True) -> None:
        """
        Remove Paragraphs by order between para_no_begin and para_no_end from the Entry
        Only removes a single Paragraph when para_no_end < para_no_begin
        """
    def set_creation_date(self, date: typing.SupportsInt | typing.SupportsIndex) -> None:
        """
        Sets the Entry's creation date
        """
    def set_text(self, text: str) -> None:
        """
        Sets the Entry's text
        Does not update Entry's edit date. Clears undo history
        """
class FT:
    """
    Types for user input fields
    
    Members:
    
      DATE
    
      CHECK
    
      OPTION
    
      TEXT
    
      FILE
    
      FOLDER
    
      TAG
    
      FILTER
    
      HIDDEN
    """
    CHECK: typing.ClassVar[FT]  # value = <FT.CHECK: 0>
    DATE: typing.ClassVar[FT]  # value = <FT.DATE: 3>
    FILE: typing.ClassVar[FT]  # value = <FT.FILE: 4>
    FILTER: typing.ClassVar[FT]  # value = <FT.FILTER: 7>
    FOLDER: typing.ClassVar[FT]  # value = <FT.FOLDER: 5>
    HIDDEN: typing.ClassVar[FT]  # value = <FT.HIDDEN: 8>
    OPTION: typing.ClassVar[FT]  # value = <FT.OPTION: 1>
    TAG: typing.ClassVar[FT]  # value = <FT.TAG: 6>
    TEXT: typing.ClassVar[FT]  # value = <FT.TEXT: 2>
    __members__: typing.ClassVar[dict[str, FT]]  # value = {'DATE': <FT.DATE: 3>, 'CHECK': <FT.CHECK: 0>, 'OPTION': <FT.OPTION: 1>, 'TEXT': <FT.TEXT: 2>, 'FILE': <FT.FILE: 4>, 'FOLDER': <FT.FOLDER: 5>, 'TAG': <FT.TAG: 6>, 'FILTER': <FT.FILTER: 7>, 'HIDDEN': <FT.HIDDEN: 8>}
    def __eq__(self, other: typing.Any) -> bool:
        ...
    def __getstate__(self) -> int:
        ...
    def __hash__(self) -> int:
        ...
    def __index__(self) -> int:
        ...
    def __init__(self, value: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def __int__(self) -> int:
        ...
    def __ne__(self, other: typing.Any) -> bool:
        ...
    def __repr__(self) -> str:
        ...
    def __setstate__(self, state: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def __str__(self) -> str:
        ...
    @property
    def name(self) -> str:
        ...
    @property
    def value(self) -> int:
        ...
class HiddenFormat:
    uri: str
    def __init__(self, arg0: typing.SupportsInt | typing.SupportsIndex, arg1: str, arg2: typing.SupportsInt | typing.SupportsIndex, arg3: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    @property
    def pos_bgn(self) -> int:
        ...
    @pos_bgn.setter
    def pos_bgn(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    @property
    def pos_end(self) -> int:
        ...
    @pos_end.setter
    def pos_end(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    @property
    def ref_id(self) -> int:
        ...
    @ref_id.setter
    def ref_id(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    @property
    def type(self) -> int:
        ...
    @type.setter
    def type(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    @property
    def var_d(self) -> int:
        ...
    @var_d.setter
    def var_d(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    @property
    def var_i(self) -> int:
        ...
    @var_i.setter
    def var_i(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
class Named:
    def get_name(self) -> str:
        ...
class ParaInhClass:
    """
    Paragraph Inheritance Class
    
    Members:
    
      QUOT_TYPE
    
      LIST_TYPE
    
      HEADING_LVL
    
      INDENTATION
    
      DEFAULT
    
      NONE
    
      SET_TEXT
    """
    DEFAULT: typing.ClassVar[ParaInhClass]  # value = <ParaInhClass.DEFAULT: 11>
    HEADING_LVL: typing.ClassVar[ParaInhClass]  # value = <ParaInhClass.HEADING_LVL: 4>
    INDENTATION: typing.ClassVar[ParaInhClass]  # value = <ParaInhClass.INDENTATION: 8>
    LIST_TYPE: typing.ClassVar[ParaInhClass]  # value = <ParaInhClass.LIST_TYPE: 2>
    NONE: typing.ClassVar[ParaInhClass]  # value = <ParaInhClass.NONE: 0>
    QUOT_TYPE: typing.ClassVar[ParaInhClass]  # value = <ParaInhClass.QUOT_TYPE: 1>
    SET_TEXT: typing.ClassVar[ParaInhClass]  # value = <ParaInhClass.SET_TEXT: 2097152>
    __members__: typing.ClassVar[dict[str, ParaInhClass]]  # value = {'QUOT_TYPE': <ParaInhClass.QUOT_TYPE: 1>, 'LIST_TYPE': <ParaInhClass.LIST_TYPE: 2>, 'HEADING_LVL': <ParaInhClass.HEADING_LVL: 4>, 'INDENTATION': <ParaInhClass.INDENTATION: 8>, 'DEFAULT': <ParaInhClass.DEFAULT: 11>, 'NONE': <ParaInhClass.NONE: 0>, 'SET_TEXT': <ParaInhClass.SET_TEXT: 2097152>}
    def __eq__(self, other: typing.Any) -> bool:
        ...
    def __ge__(self, other: typing.Any) -> bool:
        ...
    def __getstate__(self) -> int:
        ...
    def __gt__(self, other: typing.Any) -> bool:
        ...
    def __hash__(self) -> int:
        ...
    def __index__(self) -> int:
        ...
    def __init__(self, value: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def __int__(self) -> int:
        ...
    def __le__(self, other: typing.Any) -> bool:
        ...
    def __lt__(self, other: typing.Any) -> bool:
        ...
    def __ne__(self, other: typing.Any) -> bool:
        ...
    def __repr__(self) -> str:
        ...
    def __setstate__(self, state: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def __str__(self) -> str:
        ...
    @property
    def name(self) -> str:
        ...
    @property
    def value(self) -> int:
        ...
class Paragraph(DiaryElemTag):
    @typing.overload
    def add_format(self, arg0: typing.SupportsInt | typing.SupportsIndex, arg1: typing.SupportsInt | typing.SupportsIndex, arg2: typing.SupportsInt | typing.SupportsIndex) -> HiddenFormat:
        ...
    @typing.overload
    def add_format(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> HiddenFormat:
        """
        Convenience method for adding a format to the whole paragraph
        """
    def add_link(self, arg0: str, arg1: typing.SupportsInt | typing.SupportsIndex, arg2: typing.SupportsInt | typing.SupportsIndex) -> HiddenFormat:
        ...
    def add_tag(self, arg0: DiaryElemTag, arg1: typing.SupportsInt | typing.SupportsIndex) -> HiddenFormat:
        ...
    def clear_location(self) -> None:
        """
        Clears the location associated with the paragraph if any
        """
    def erase_text(self, pos_bgn: typing.SupportsInt | typing.SupportsIndex, len_erase: typing.SupportsInt | typing.SupportsIndex, add_undo: bool = True) -> None:
        ...
    def get_alignment(self) -> int:
        ...
    def get_code_lang(self) -> str:
        ...
    def get_format_at(self, arg0: typing.SupportsInt | typing.SupportsIndex, arg1: typing.SupportsInt | typing.SupportsIndex) -> HiddenFormat:
        ...
    def get_formats(self) -> list[HiddenFormat]:
        ...
    def get_heading_level(self) -> int:
        ...
    def get_host(self) -> Entry:
        """
        Get the host Entry
        """
    def get_image_size(self) -> int:
        ...
    def get_indent_level(self) -> int:
        ...
    def get_last_sub(self) -> Paragraph:
        ...
    def get_latitude(self) -> float:
        ...
    def get_list_type(self) -> int:
        ...
    def get_longitude(self) -> float:
        ...
    def get_next(self) -> Paragraph:
        ...
    def get_next_sibling(self) -> Paragraph:
        ...
    def get_parent(self) -> Paragraph:
        ...
    def get_prev(self) -> Paragraph:
        ...
    def get_prev_sibling(self) -> Paragraph:
        ...
    def get_quot_type(self) -> str:
        ...
    def get_subtag_first(self, arg0: DiaryElemTag) -> DiaryElemTag:
        ...
    def get_subtag_highest(self, arg0: DiaryElemTag) -> DiaryElemTag:
        ...
    def get_subtag_last(self, arg0: DiaryElemTag) -> DiaryElemTag:
        ...
    def get_subtag_lowest(self, arg0: DiaryElemTag) -> DiaryElemTag:
        ...
    def get_tag_def_bound(self) -> int:
        ...
    def get_text(self) -> str:
        ...
    def get_uri(self) -> str:
        ...
    def has_hrule(self) -> bool:
        ...
    def has_location(self) -> bool:
        """
        Returns True if the Paragraph has map location
        """
    def insert_text(self, pos: typing.SupportsInt | typing.SupportsIndex, text: str, add_undo: bool = True) -> None:
        ...
    def is_code(self) -> bool:
        ...
    def is_empty_completely(self, arg0: bool) -> bool:
        ...
    def is_expanded(self) -> bool:
        ...
    def is_image(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> bool:
        ...
    def is_quote(self) -> bool:
        ...
    def move_to_entry(self, new_host: Entry, para_before: Paragraph = None) -> None:
        """
        Move Paragraph to a different Entry after para_before. Appends to end if para_before is None.
         Clears undo history of the source Entry
        """
    def remove_format(self, arg0: HiddenFormat) -> None:
        ...
    def replace_text(self, pos_bgn: typing.SupportsInt | typing.SupportsIndex, len_erase: typing.SupportsInt | typing.SupportsIndex, new_text: str, add_undo: bool = True) -> None:
        ...
    def set_alignment(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def set_expanded(self, arg0: bool) -> None:
        ...
    def set_heading_level(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def set_hrule(self, arg0: bool) -> None:
        ...
    def set_image_size(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        """
        Image size has to be an integer between [0,3]
        """
    def set_image_type(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        """
        Set the image type as FILE, CHART or TABLE. Setting to 0 clears
        """
    def set_indent_level(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> bool:
        ...
    def set_list_type(self, type: typing.SupportsInt | typing.SupportsIndex) -> None:
        """
        Set list type from the options in VT.PLS.*
        """
    def set_location(self, arg0: typing.SupportsFloat | typing.SupportsIndex, arg1: typing.SupportsFloat | typing.SupportsIndex) -> None:
        """
        Set map location of a Paragraph as latitude and longitude
        """
    def set_quot_type(self, type: str) -> None:
        """
        Set quotation type from the options in VT.QT.*
        """
    def set_quote(self, is_quote: bool = True) -> None:
        """
        Set to literary quote
        """
    def set_tag_def_bound(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def set_text(self, text_new: str, add_undo: bool = True) -> None:
        ...
    def set_uri(self, arg0: str) -> None:
        ...
class ParserBackGround:
    def __init__(self, arg0: Diary) -> None:
        ...
class SCtg:
    """
    Members:
    
      TBD
    
      NONE
    
      IMPORTER
    
      EXPORTER
    
      TABLE_COLUMN
    
      WORD_MODIFIER
    
      PARA_MODIFIER
    
      ENTRY_MODIFIER
    
      FILTER
    """
    ENTRY_MODIFIER: typing.ClassVar[SCtg]  # value = <SCtg.ENTRY_MODIFIER: 7>
    EXPORTER: typing.ClassVar[SCtg]  # value = <SCtg.EXPORTER: 3>
    FILTER: typing.ClassVar[SCtg]  # value = <SCtg.FILTER: 8>
    IMPORTER: typing.ClassVar[SCtg]  # value = <SCtg.IMPORTER: 2>
    NONE: typing.ClassVar[SCtg]  # value = <SCtg.NONE: 1>
    PARA_MODIFIER: typing.ClassVar[SCtg]  # value = <SCtg.PARA_MODIFIER: 6>
    TABLE_COLUMN: typing.ClassVar[SCtg]  # value = <SCtg.TABLE_COLUMN: 4>
    TBD: typing.ClassVar[SCtg]  # value = <SCtg.TBD: 0>
    WORD_MODIFIER: typing.ClassVar[SCtg]  # value = <SCtg.WORD_MODIFIER: 5>
    __members__: typing.ClassVar[dict[str, SCtg]]  # value = {'TBD': <SCtg.TBD: 0>, 'NONE': <SCtg.NONE: 1>, 'IMPORTER': <SCtg.IMPORTER: 2>, 'EXPORTER': <SCtg.EXPORTER: 3>, 'TABLE_COLUMN': <SCtg.TABLE_COLUMN: 4>, 'WORD_MODIFIER': <SCtg.WORD_MODIFIER: 5>, 'PARA_MODIFIER': <SCtg.PARA_MODIFIER: 6>, 'ENTRY_MODIFIER': <SCtg.ENTRY_MODIFIER: 7>, 'FILTER': <SCtg.FILTER: 8>}
    def __eq__(self, other: typing.Any) -> bool:
        ...
    def __getstate__(self) -> int:
        ...
    def __hash__(self) -> int:
        ...
    def __index__(self) -> int:
        ...
    def __init__(self, value: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def __int__(self) -> int:
        ...
    def __ne__(self, other: typing.Any) -> bool:
        ...
    def __repr__(self) -> str:
        ...
    def __setstate__(self, state: typing.SupportsInt | typing.SupportsIndex) -> None:
        ...
    def __str__(self) -> str:
        ...
    @property
    def name(self) -> str:
        ...
    @property
    def value(self) -> int:
        ...
class TableColumn(Named):
    def format_number(self, arg0: TableLine) -> None:
        ...
    def get_index(self) -> int:
        ...
class TableLine:
    def get_elem(self) -> DiaryElemTag:
        ...
    def get_value_num(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> float:
        ...
    def get_value_txt(self, arg0: typing.SupportsInt | typing.SupportsIndex) -> str:
        ...
    def set_value_num(self, arg0: typing.SupportsInt | typing.SupportsIndex, arg1: typing.SupportsFloat | typing.SupportsIndex, arg2: typing.SupportsFloat | typing.SupportsIndex) -> None:
        ...
    def set_value_txt(self, arg0: typing.SupportsInt | typing.SupportsIndex, arg1: str) -> None:
        ...
def get_user_input(*args, **kwargs) -> None:
    """
    Displays a dialog for collecting user input and executes a script upon completion
    Parameters:
    info: Descriptive test to be shown on the dialog
    name: Name of the script to call upon completion
    cb_name: Name of the callback funtion in that script
    fields: List of input fields as a fields in the form of FTs
    """
def refresh_panels() -> None:
    """
    Refreshes text editor, entry list, and the extra panel to reflect changes
    """
def show_element(id: typing.SupportsInt | typing.SupportsIndex) -> None:
    """
    Shows the DiaryElement with the id if it exists (Works for Entries, Paragraphs, Charts, and Tables)
    """
def show_message(message: str) -> None:
    """
    Displays the given message in the info bar
    """
def update_for_new_entry(entry: Entry) -> None:
    """
    Adds entry to the list and displays it
    """



# Copyright (C) 2007-2026 Ahmet Öztürk (aoz_2@yahoo.com)
#
# This file is part of Lifeograph.
#
# Lifeograph is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lifeograph is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Lifeograph.  If not, see <http://www.gnu.org/licenses/>.

from LoG import *
from pathlib import Path
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt

NAME="Render Math Formula (Experimental)"
VERSION="0.1"
CATEGORY=SCtg.WORD_MODIFIER

def run(para: Paragraph, pos_bgn, pos_end):
    img_uri = _render_formula_for_para(para)
    para.set_uri(img_uri)
    para.set_image_type(VT.PS.IMAGE_FILE)
    para.set_image_size(0) # minimum
    para.set_alignment(VT.PS.ALIGN_C)
    para.add_format(VT.FMT.FADED)
    refresh_panels()

def _render_formula_for_para(para) -> str:
    formula = para.get_text().strip()
    para_id = para.get_id()

    base_dir = Path(Diary.get_main().get_rel_folder_path())
    formulas_dir = base_dir / "formulas"
    formulas_dir.mkdir(parents=True, exist_ok=True)
    filename = f"{para_id}.png"

    out_path = formulas_dir / filename
    img_rel_uri = (
        "rel://" +
        Diary.get_main().get_rel_folder_name() +
        "/formulas/" +
        filename )

    fig = plt.figure(figsize=(0.01, 0.01))
    fig.patch.set_alpha(0)

    plt.text(
        0.5, 0.5,
        f"${formula}$",
        fontsize=18,
        ha="center",
        va="center"
    )

    plt.axis("off")

    plt.savefig(
        out_path,
        format="png",
        dpi=300,
        bbox_inches="tight",
        pad_inches=0.05
    )
    plt.close(fig)

    return img_rel_uri
